import * as d3geo from 'd3-geo';
import * as d3geoProjection from 'd3-geo-projection';
import * as d3geoPolygon from 'd3-geo-polygon';
import { PNG } from 'pngjs';
import { readFile } from 'fs/promises';

type Coords = [number, number]
type Bounds = [Coords, Coords]

type ProjectionFunction = (...args: unknown[]) => d3geo.GeoProjection;

const referenceWidth = 960 as const;
const referenceHeight = 480 as const;

type ProjectOptions = {
  inputDimensions?: { width: number, height: number },
  bounds?: Bounds,
  wrapAround?: boolean,
}

type Projection = [string, ...unknown[]]

const getCenter = (bounds: Bounds): Coords => {
  const [nw, se] = bounds;
  const [nwLng, nwLat] = nw;
  const [seLng, seLat] = se;
  return [
    // TODO: better center checking that wraps longitudes
    (nwLng + seLng) / 2,
    (nwLat + seLat) / 2,
  ];
};

const getProjectionFunctionSource = (projectionFunctionName: string): Record<string, unknown> => {
  if (projectionFunctionName in d3geoPolygon) {
    return d3geoPolygon as Record<string, unknown>;
  }
  if (projectionFunctionName in d3geoProjection) {
    return d3geoProjection as Record<string, unknown>;
  }
  if (projectionFunctionName in d3geo) {
    return d3geo as Record<string, unknown>;
  }

  throw new Error(`Unknown projection: ${projectionFunctionName}`);
};

const buildProjectionFunction = (projection: Projection, options: ProjectOptions = {}) => {
  const [projectionName, ...projectionArgs] = projection;
  const projectionFunctionName = `geo${projectionName}`;
  const geoProjectionSource = getProjectionFunctionSource(projectionFunctionName);
  const projectionFn = geoProjectionSource[projectionFunctionName] as ProjectionFunction;
  const baseProjection = projectionFn(...projectionArgs) as unknown as d3geo.GeoProjection;
  const {
    bounds = [[-180, 90], [180, -90]],
  } = options;

  return baseProjection.center(getCenter(bounds));
};

export const projectPoint = (
  point: [number, number],
  projection: Projection,
  options: ProjectOptions = {},
): [number, number] | null => {
  const [x, y] = point;

  const [projectionName] = projection;
  if (projectionName === 'Equirectangular') {
    return point;
  }

  const {
    inputDimensions = { width: referenceWidth, height: referenceHeight },
    wrapAround = false,
  } = options;

  const projectionFunction = buildProjectionFunction(projection, options);
  if (!projectionFunction.invert) {
    return null;
  }

  const { width: sx, height: sy } = inputDimensions;
  const originalPoint = [
    (x / sx) * referenceWidth,
    (y / sy) * referenceHeight,
  ] as [number, number];
  const inverted = projectionFunction.invert(originalPoint);

  if (!inverted) {
    return null;
  }

  if (wrapAround) {
    return inverted;
  }

  const validation = projectionFunction(inverted);
  if (!validation) {
    return null;
  }

  // https://stackoverflow.com/questions/41832043/how-to-fix-map-boundaries-on-d3-cartographic-raster-reprojection/41856996#41856996
  const tolerance = 0.5;

  if (
    Math.abs(validation[0] - originalPoint[0]) < tolerance
    && Math.abs(validation[1] - originalPoint[1]) < tolerance
  ) {
    // to avoid wrapping, let's only pick the points that inverts back to its original point
    return inverted;
  }

  return null;
};

const retrievePngData = async (pngInput: string | Buffer | PNG) => {
  if (typeof pngInput === 'string') {
    const pngFile = await readFile(pngInput);
    return PNG.sync.read(pngFile);
  }

  if (typeof pngInput === 'object') {
    return pngInput instanceof PNG ? pngInput : PNG.sync.read(pngInput);
  }

  throw new TypeError('Invalid input argument.');
};

const prepareOutputData = (
  projection: Projection,
  sourcePng: { width: number, height: number },
  options = {} as ProjectOptions,
): PNG => {
  const p = buildProjectionFunction(projection, options);
  const {
    bounds = [[-180, 90], [180, -90]],
  } = options;
  const nwBoundsRaw = p(bounds[0]);
  const seBoundsRaw = p(bounds[1]);
  const nwBounds = nwBoundsRaw ? [
    (nwBoundsRaw[0] / referenceWidth) * sourcePng.width,
    (nwBoundsRaw[1] / referenceHeight) * sourcePng.height,
  ] : null;

  const seBounds = seBoundsRaw ? [
    (seBoundsRaw[0] / referenceWidth) * sourcePng.width,
    (seBoundsRaw[1] / referenceHeight) * sourcePng.height,
  ] : null;

  let tx = sourcePng.width;
  let ty = sourcePng.height;
  if (seBounds && nwBounds) {
    tx = (nwBounds[0] + seBounds[0]);
    ty = (nwBounds[1] + seBounds[1]);

    if (
      projection[0] === 'Mercator'
      && ty > tx
    ) {
      ty = tx;
    }
  }

  return new PNG({
    width: tx,
    height: ty,
  });
};

export const project = async (
  pngInput: string | Buffer | PNG,
  projection: Projection,
  options?: ProjectOptions,
) => {
  const sourcePngData = await retrievePngData(pngInput);
  const { width: sx, height: sy, data: sourceData } = sourcePngData;
  const outputPngData = prepareOutputData(projection, sourcePngData, options);
  const { width: tx, height: ty, data: targetData } = outputPngData;

  let i = 0;
  for (let y = (sy - ty) / 2; y < (ty + ((sy - ty) / 2)); y += 1) {
    for (let x = 0; x < (tx + ((sx - tx) / 2)); x += 1) {
      const projected = projectPoint([x, y], projection, {
        bounds: options?.bounds,
        inputDimensions: {
          width: sx,
          height: sy,
        },
      });
      if (projected) {
        const [lambda, phi] = projected;
        if (!(lambda > 180 || lambda < -180 || phi > 90 || phi < -90)) {
          // eslint-disable-next-line no-bitwise
          const q = (((90 - phi) / 180) * sy | 0) * sx + (((180 + lambda) / 360) * sx | 0) << 2;
          targetData[i] = sourceData[q];
          targetData[i + 1] = sourceData[q + 1];
          targetData[i + 2] = sourceData[q + 2];
          targetData[i + 3] = sourceData[q + 3];
        }
      }
      i += 4;
    }
  }

  return PNG.sync.write(outputPngData);
};
